import inspect
import argparse

def default_parser(func, configs=None, **kwargs):
    if configs is None:
        configs = {}

    signature = inspect.signature(func)
    
    parser = argparse.ArgumentParser(**kwargs)
    for k, v in signature.parameters.items():
        if k in configs and 'opt_str' in configs[k]:
            option_strings_k = configs[k]['opt_str']
        else:
            if v.default is inspect.Parameter.empty:
                option_strings_k = []
            else:
                if len(k) == 1:
                    option_strings_k = [f'-{k}'.replace('_','-')]
                else:
                    option_strings_k = [f'--{k}'.replace('_','-')]
                
        if k in configs and 'type' in configs[k]:
            type_k = configs[k]['type']
        else:
            if v.annotation is inspect.Parameter.empty:
                type_k = str
            else:
                type_k = v.annotation
                interpretable_types = [str, int, float, bool]
                if not any(type_k == t for t in interpretable_types):
                    raise ValueError(f"type {type_k} for argument {k} is not recognized. Please provide the type manually")
                
        if k in configs and 'nargs' in configs[k]:
            nargs_k = configs[k]['nargs']
        else:
            nargs_k = None
            
        if k in configs and 'default' in configs[k]:
            default_k = configs[k]['default']
        else:
            if v.default is inspect.Parameter.empty:
                default_k = None
            else:
                default_k = v.default
                
        if k in configs and 'choices' in configs[k]:
            choices_k = configs[k]['choices']
        else:
            choices_k = None
                
        if k in configs and 'help' in configs[k]:
            help_k = configs[k]['help']
        else:
            help_k = k.replace('_', ' ')
            
        if k in configs and 'action' in configs[k]:
            action_k = configs[k]['action']
            if action_k in ['store_true', 'store_false']:
                parser.add_argument(*option_strings_k, dest=k, action=action_k, help=help_k)
            else:
                raise NotImplementedError()
        else:
            help_k = f'{help_k} (dtype: {type_k.__name__})'

            if v.default is not inspect.Parameter.empty:
                help_k = f'{help_k} (default: {default_k})'
            else:
                help_k = f'{help_k} (required argument)'
                
            if v.default is inspect.Parameter.empty and len(option_strings_k) > 0:
                parser.add_argument(*option_strings_k, dest=k, type=type_k, nargs=nargs_k, default=default_k, choices=choices_k, help=help_k, required=True)
            else:
                parser.add_argument(*option_strings_k, dest=k, type=type_k, nargs=nargs_k, default=default_k, choices=choices_k, help=help_k)
        
    return parser

def multi_parse(parsers, **kwargs):
    # IMPORTANT: Assumes that the __call__ method of each action of each parser
    # indeed stores the parsed input at the action's DEST.
    parser = argparse.ArgumentParser(parents=parsers)
    args = parser.parse_args(**kwargs)
    out = []
    for parser in parsers:
        out.append(argparse.Namespace(**{action.dest: getattr(args, action.dest) for action in parser._actions if hasattr(args, action.dest)}))
    return out


def pformat(d, indent=0, spaces=4, verbose=False):
    output = ''
    for key, value in d.items():
        output += f"{' ' * spaces * indent}'{str(key)}':\n"
        if isinstance(value, dict):
            output += f"{pformat(value, indent=indent+1, spaces=spaces, verbose=verbose)}\n"
        else:
            if not verbose:
                value = type(value)
            output += '\n'.join([f"{' ' * spaces * (indent+1)}{line}" for line in str(value).split('\n')]) + '\n'
    return output.rstrip('\n')

def pprint(d, **kwargs):
    print(pformat(d, **kwargs))


import json
import configparser
from importlib import resources
import pathlib
import logging
import os

import platformdirs
import numpy as np
import tomli

from . import itertools

logger = logging.getLogger(__name__)

def load(filename):
    filename = str(filename)
    if filename.endswith('.toml'):
        with open(filename, "rb") as f:
            config = tomli.load(f)
    elif filename.endswith('.json'):
        with open(filename, 'r') as f:
            config = json.load(f)
    else:
        raise NotImplementedError()
    
    return config

def dump(config, filename):
    filename = str(filename)
    if filename.endswith('.json'):
        with open(filename, 'w') as f:
            json.dump(config, f, indent=4)
    else:
        raise NotImplementedError()
        
def package_config_locs(package_name, package_author=None, package_version=None, default_config_path='config.toml'):
    default_config_filename = resources.files(package_name).joinpath(default_config_path)
    
    user_config_filenames = []

    user_config_filenames.append(pathlib.Path(f'{package_name}_config.toml'))
    try:
        path = pathlib.Path(os.environ[f'{package_name.upper()}_CONFIG'])
    except KeyError:
        pass
    else:
        user_config_filenames.append(path if path.is_file() else path / f'{package_name}_config.toml')
    user_config_filenames.append(pathlib.Path(
        platformdirs.user_config_dir(package_name, appauthor=package_author, version=package_version)
    ) / 'config.toml')
    
    return {'default_config': default_config_filename, 'user_configs': user_config_filenames}
        
def load_package_config(*args, **kwargs):
    """
    Loads configs from various config files to be imported and used anywhere in a package.
    This is meant to be used in the top-level __init__.py file in the package.
    
    It first loads default package configs at default_config_path 
    (relative to the top level directory of the package) if such a file exists.
    
    Next, it loads user-defined configs in the following priority, from highest to lowest:
        1. f'{package_name}_config.toml' in the directory in which the top level code is run
        2. f'${package_name.upper()}_CONIFG' if it is a file, f'${package_name.upper()}_CONIFG/{package_name}_config.toml' otherwise
        3. f'{platformdirs.user_config_dir(package_name, appauthor=package_author, version=package_version)}/config.toml'
        
    If there are overlapping configs, the configs from the config file with the higher priority will be used.
    
    Reminder: If you want to add a default config file to your package, remember to set include_package_data=True in setup.py
    to include it in your package distribution.
    """
    filenames = package_config_locs(*args, **kwargs)
    
    default_config_filename = filenames['default_config']
    if default_config_filename.is_file():
        with resources.as_file(default_config_filename) as default_config_file:
            config = load(default_config_file)
        
        logger.debug(f"Loaded default config file at {default_config_filename}.")
    else:
        config = {}

    user_config_filenames = [filename for filename in filenames['user_configs'] if filename.is_file()]
    for filename in reversed(user_config_filenames):
        itertools.dict_update(config, load(filename))

    if len(user_config_filenames) > 0:
        logger.debug(f"Loaded user config files at the following locations from highest to lowest priority: {user_config_filenames}")

    return config
        
def _replace_range_by_list(d, exclude=None):
    if exclude is None:
        exclude = []
    
    for k, v in d.items():
        if isinstance(v, dict) and k not in exclude:
            if any([vk not in ['start', 'stop', 'step'] for vk in v.keys()]):
                raise ValueError('configuration values cannot be dictionaries, unless it represents a range')
            d[k] = np.arange(**v).tolist()

def expand(d):
    _replace_range_by_list(d, exclude=['zip', 'prod'])
    
    configs = []
    
    e = {k: v for k, v in d.items() if k not in ['zip', 'prod']}
    if len(e) > 0:
        configs.append(e)
    
    if 'zip' in d:
        l = d['zip']
        if isinstance(l, dict):
            l = [l]
        assert isinstance(l, list)

        for e in l:
            assert isinstance(e, dict) and len(e) > 0
            _replace_range_by_list(e)

            Ns = [len(v) for v in e.values()]
            N = Ns[0]
            assert all([N == Ns[0] for N in Ns])

            configs += [{k: v[i] for k, v in e.items()} for i in range(N)]
    
    if 'prod' in d:
        l = d['prod']
        if isinstance(l, dict):
            l = [l]
        assert isinstance(l, list)

        for e in l:
            assert isinstance(e, dict)
            _replace_range_by_list(e)
            
            for k, v in e.items():
                if not isinstance(v, list):
                    e[k] = [v]

            ks, vs = zip(*list(e.items())) # gauranteed ks and vs are in same order
            shape = tuple([len(v) for v in vs])
            for indices in np.ndindex(shape):
                configs.append({k: v[i] for k, v, i in zip(ks, vs, indices)})
            
    return configs


class PathNotFound(IOError):
    pass

class PathAlreadyExists(IOError):
     pass

class ConfigurableError(Exception):
    pass

class InvalidConfigParameter(ConfigurableError):
    pass

class ConfigDictError(ConfigurableError):
    pass


import numbers
import functools

import numpy as np

class ndarray(np.ndarray):
    def __array_finalize__(self, obj):
        if obj is None: return
        self.op_expr = getattr(obj, 'op_expr', '{}')
    
    def __call__(self, x, batch='leading'):
        if isinstance(x, numbers.Number):
            x_shape = tuple()
            x = np.array([x])
        elif isinstance(x, np.ndarray):
            x_shape = x.shape
            x = x.reshape(-1)
        else:
            raise NotImplementedError(f"Currently only accepts scalar or np.ndarray inputs, but got {type(x)}")
        
        res = []
        for idx in np.ndindex(self.shape):
            if not callable(self[idx]):
                res.append(self[idx] + x*0)
            else:
                res.append(self[idx](x))
                
        if batch == 'leading':
            res = np.moveaxis(np.array(res), -1, 0).reshape((*x_shape, *self.shape))
        elif batch == 'trailing':
            res = np.array(res).reshape((*self.shape, *x_shape))
        else:
            raise ValueError(f"batch argument must be 'leading' or 'trailing', but '{batch}' provided.")
            
        return res
    
    def __matmul__(self, g):
        result = matmul(self, g)
        result.op_expr = '({} @ {})'
        return result
    
    def __repr__(self):
        op_expr = self.op_expr.format(*[f'f{i}' for i in range(self.op_expr.count('{}'))])
        return f"npf.ndarray(shape={self.shape}, op_expr={op_expr})"
    
    def __str__(self):
        return repr(self)
    
def array(*args, **kwargs):
    arr = np.array(*args, **kwargs).view(ndarray)
    return arr

def empty(*args, dtype=object, **kwargs):
    arr = np.empty(*args, dtype=dtype, **kwargs).view(ndarray)
    return arr

### npf.ndarray ufuncs ###

def sum(x, **kwargs):
    x = array(x)
    add_ufunc = get_ufunc('__add__')
    result = add_ufunc.reduce(x, **kwargs)
    result.op_expr = f'sum{x.op_expr}'
    return result

@functools.partial(np.vectorize, signature='(a,b),(b,c)->(a,c)')
def _matmul(x, y):
    assert x.ndim == 2 and y.ndim == 2 and x.shape[1] == y.shape[0]
    z = empty((x.shape[0], y.shape[1]))
    for i, j in np.ndindex(z.shape):
        z[i,j] = sum(x[i,:] * y[:,j])
    return z

def matmul(x, y):
    squeeze_x, squeeze_y = False, False
    if x.ndim == 1:
        x = x[None,:]
        squeeze_x = True
    if y.ndim == 1:
        y = y[:,None]
        squeeze_y = True
    result = _matmul(x, y)
    if squeeze_x:
        result = result.squeeze(-2)
    if squeeze_y:
        result = result.squeeze(-1)
    return result

def trace(A):
    return sum(np.diagonal(A, axis1=-2, axis2=-1))

@functools.partial(np.vectorize, signature='(m,n)->()')
def _det(A):
    assert A.ndim == 2 and A.shape[0] == A.shape[1], f'A ({shape=}) is not a square matrix'
    N = A.shape[0]
    
    if N == 1:
        return array(A.item())

    res = sum([((-1)**i * array(A[0,i]) * _minor(A, 0, i)).item() for i in range(N)])

    return res

def det(A):
    """
    The reason I need to do this is because of probably some bug with np.vectorize
    that causes _det to return array(array(obj)) rather than array(obj).
    I don't even know how it does that, for example actually explicitly doing
    utils.npf.array(utils.npf.array(lambda x: x)) returns utils.npf.array(lambda x: x).
    This gaurantees return of utils.npf.array(obj).
    """
    return array(_det(A).item())

def _minor(A, i, j):
    assert A.ndim == 2 and A.shape[0] == A.shape[1], f'A ({shape=}) is not a square matrix'

    res = det(np.delete(np.delete(A, i, axis=0), j, axis=1))

    return res

@functools.partial(np.vectorize, signature='(m,n)->(m,n)')
def adj(A):
    assert A.ndim == 2 and A.shape[0] == A.shape[1], f'A ({shape=}) is not a square matrix'
    N = A.shape[0]
    return array([[((-1)**(i+j) * _minor(A, i, j)).item() for j in range(N)] for i in range(N)]).T

def inv(A):
    return adj(A)/det(A)

def get_ufunc(op_name):
    assert op_name in OP_EXPR_DICT
    rop_name = '__r' + op_name[2:]
    if rop_name not in OP_EXPR_DICT:
        rop_name = None
        
    def ufunc(f, g):
        if hasattr(f, op_name):
            h = getattr(f, op_name)(g)
            if h is NotImplemented:
                if rop_name is not None:
                    if hasattr(g, rop_name):
                        h = getattr(g, rop_name)(f)
                    elif callable(g):
                        if isinstance(f, numbers.Number):
                            h = lambda x: getattr(g(x), rop_name)(f)
                        elif callable(f):
                            h = lambda x: getattr(g(x), rop_name)(f(x))
                        else:
                            return NotImplemented
                    else:
                        return NotImplemented
                else:
                    return NotImplemented
        elif callable(f):
            if isinstance(g, numbers.Number):
                h = lambda x: getattr(f(x), op_name)(g)
            elif callable(g):
                h = lambda x: getattr(f(x), op_name)(g(x))
            else:
                return NotImplemented
        else:
            return NotImplemented
        return h
    
    return np.frompyfunc(ufunc,2,1)

### add methods for common mathematical operators to npf.ndarray ###

OP_EXPR_DICT = {
    '__add__': '+',
    '__sub__': '-',
    '__mul__': '*',
    '__truediv__': '/',
    '__pow__': '**',
    '__radd__': '+',
    '__rsub__': '-',
    '__rmul__': '*',
    '__rtruediv__': '/',
    '__rpow__': '**',
}

def get_array_op(op_name):
    ufunc = get_ufunc(op_name)
    
    def array_op(self, g):
        try:
            result = ufunc(self, g)
        except TypeError:
            return NotImplemented
        
        if np.any(result == NotImplemented):
            return NotImplemented
        
        if isinstance(g, ndarray):
            result.op_expr = f'({self.op_expr} {OP_EXPR_DICT[op_name]} {g.op_expr})'
        else:
            if op_name.startswith('__r'):
                result.op_expr = f'({g} {OP_EXPR_DICT[op_name]} {self.op_expr})'
            else:
                result.op_expr = f'({self.op_expr} {OP_EXPR_DICT[op_name]} {g})'
        return result
    
    return array_op

for op_name in OP_EXPR_DICT.keys():
    setattr(ndarray, op_name, get_array_op(op_name))


import itertools

import numpy as np

def flatten_seq(s, depth=-1, dtypes=[list, tuple]):
    """
    Recursively flattens a sequence (defined as an instance
    of one of the DTYPES) up to depth DEPTH.
    Will try to cast the result to type(s) if possible, otherwise returns a list.
    depth=0 returns the sequence unchanged, depth=-1 returns a fully flattened sequence
    TypeError is raised if s is not an instance of any of the dtypes.
    
    >>> s = [[1,2,[3]],[],(4,5),6]
    >>> flatten_seq(s, depth=1)
    [1,2,[3],4,5,6]
    >>> flatten_seq(s, depth=-1)
    [1,2,3,4,5,6]
    >>> flatten_seq(tuple(s), depth=-1)
    (1,2,3,4,5,6)
    
    """
    if not any([isinstance(s, dtype) for dtype in dtypes]):
        raise TypeError(f"The arugment to flatten_list must be one of {dtypes}, not {type(s)}")
    flattened_s = []
    for elem in s:
        if depth == 0 or not any([isinstance(elem, dtype) for dtype in dtypes]):
            flattened_s.append(elem)
        else:
            flattened_s += flatten_seq(elem, depth=depth-1, dtypes=dtypes)
    try:
        return type(s)(flattened_s)
    except Exception:
        return flattened_s
    
def flatten_dict(d, depth=-1):
    """
    Recursively flattens dict up to depth DEPTH. Keys are concatenated into a tuple key.
    depth=0 returns the dictionary unchanged, depth=-1 returns a fully flattened dictionary
    depth=n means the first n layers of the dictionary is flattened, so a depth k dict becomes a depth k-1 dict.
    TypeError is raised if d is not a dict.
    
    >>> d = {'a': {'b': {'c': 'd'}}}
    >>> flatten_dict(d, depth=1)
    {'a.b': {'c': 'd'}}
    >>> flatten_dict(d, depth=-1)
    {'a.b.c':'d'}
    
    """
    if not isinstance(d, dict):
        raise TypeError(f"The arugment to flatten_dict must be a dict, not {type(d)}")
    flattened_dict = {}
    for k, v in d.items():
        if '.' in k:
            raise ValueError("dictionary should not contain '.' in any of its keys")
        if depth == 0 or not isinstance(v, dict) or v == {}: # if depth == 0 or v is leaf
            flattened_dict[k] = v
        else:
            for new_k, new_v in flatten_dict(v, depth=depth-1).items():
                flattened_dict['.'.join([k, new_k])] = new_v
    return flattened_dict

def dict_iter(d, delimiter='.'):
    """
    Iterates over all leaf nodes of a dictionary in (key, value) pairs
    If delimiter is None, key is a tuple of the nested keys
    Otherwise, key is a string with the nested keys joined by delimiter
    """
    assert isinstance(d, dict)
    
    for k, v in d.items():
        if isinstance(v, dict):
            for ki, vi in dict_iter(v, delimiter=delimiter):
                if delimiter is None:
                    yield ((k, *ki), vi)
                else:
                    yield (delimiter.join([k, ki]), vi)
        
        else:
            if delimiter is None:
                yield ((k,), v)
            else:
                yield (k, v)
            
def dict_get(d, k, delimiter='.'):
    if delimiter is None:
        ks = k
    else:
        ks = k.split(delimiter)
        
    for ki in ks:
        d = d[ki]
    return d

# more general version of assign_dict
def dict_set(d, k, v, delimiter='.'):
    if delimiter is None:
        ks = k
    else:
        ks = k.split(delimiter)
        
    for ki in ks[:-1]:
        d = d.setdefault(ki, {})
    d[ks[-1]] = v
    
def dict_update(d, new_d):
    for k, v in dict_iter(new_d, delimiter=None): # might get very slow if new_d is very large, but I can't think of a better way
        dict_set(d, k, v, delimiter=None)
        
def dict_union(d, new_d):
    d = d.copy()
    dict_update(d, new_d)
    return d

def assign_dict(d, keys, value):
    """
    Assign (potentially nested) value to dictionary using a list/tuple of keys
    This is an in-place operation.
    
    >>> d = {}
    >>> assign_dict(d, ['a', 'b'], 'c')
    >>> d
    {'a': {'b': 'c'}}
    
    """
    for key in keys[:-1]:
        d = d.setdefault(key, {})
    d[keys[-1]] = value
    
def product(*ls, enum=False):
    """
    Like itertools.product, but has an enum keyword that
    returns the nd-index along with the elements. Note, however, it
    assumes that the inputs are sequences, not iterables, since len() is used.
    """
    if not enum:
        return itertools.product(*ls)
    return zip(np.ndindex(tuple(len(l) for l in ls)), itertools.product(*ls))
    
def dict_zip(*dicts, mode='strict', **kwargs):
    if mode == 'strict':
        return _dict_zip_strict(*dicts, **kwargs)
    if mode == 'intersect':
        return _dict_zip_intersection(*dicts, **kwargs)
    if mode == 'union':
        return _dict_zip_union(*dicts, **kwargs)
    raise NotImplementedError("mode must be either 'strict', 'intersect', or 'union'")
            
def _dict_zip_strict(*dicts):
    if not dicts:
        return

    n = len(dicts[0])
    if any(len(d) != n for d in dicts):
        raise ValueError('arguments must have the same length')

    for key, first_val in dicts[0].items():
        yield key, first_val, *(other[key] for other in dicts[1:])


def _dict_zip_intersection(*dicts):
    if not dicts:
        return

    keys = set(dicts[0]).intersection(*dicts[1:])
    for key in keys:
        yield key, *(d[key] for d in dicts)


def _dict_zip_union(*dicts, fillvalue=None):
    if not dicts:
        return

    keys = set(dicts[0]).union(*dicts[1:])
    for key in keys:
        yield key, *(d.get(key, fillvalue) for d in dicts)



import pathlib
import signal
import sys
import logging

logger = logging.getLogger(__name__)

#########################################################
### python script containing various context managers ###
#########################################################

class SimulFileHandler:
    """
    A context manager that ensures all the files asscoiated with filenames 
    must coexist. If any one file is missing upon exiting the with block
    (either because of an exception, a SIGTERM, or normal exiting),
    all the files will be deleted.
    """
    def __init__(self, *filenames):
        self.filenames = [pathlib.Path(filename) for filename in filenames]
        
    def cleanup(self):
        if all([filename.is_file() for filename in self.filenames]):
            logger.debug("All files exist. No cleanup needed.")
            return
        logger.debug("Cleaning up...")
        for filename in self.filenames:
            if filename.is_file():
                filename.unlink()
        logger.debug("Finished cleanup.")
        
    def handler(self, signum, frame):
        logger.info(f"Received signal {signal.strsignal(signum)}.")
        sys.exit(0) # This throws the exception SystemExit, which is then caught by __exit__ and triggers self.cleanup()
    
    def __enter__(self):
        self.old_sigterm = signal.signal(signal.SIGTERM, self.handler)
        
    def __exit__(self, exc_type, exc_val, exc_tb):
        if exc_type is not None:
            logger.info(f"Caught exception {exc_type}: {exc_val}")
        self.cleanup()
        signal.signal(signal.SIGTERM, self.old_sigterm)



import re
import ast

allowed_nodes = (
    ast.Constant,
    ast.List,
    ast.Tuple,
    ast.Set,
    ast.Dict,
    ast.Name,
    ast.Load,
    ast.UnaryOp,
    ast.UAdd,
    ast.USub,
    ast.Not,
    ast.Invert,
    ast.BinOp,
    ast.Add,
    ast.Sub,
    ast.Mult,
    ast.Div,
    ast.FloorDiv,
    ast.Mod,
    ast.Pow,
    ast.BitOr,
    ast.BitAnd,
    ast.BoolOp,
    ast.And,
    ast.Or,
    ast.Compare,
    ast.Eq,
    ast.NotEq,
    ast.Lt,
    ast.LtE,
    ast.Gt,
    ast.GtE,
    ast.Is,
    ast.IsNot,
    ast.In,
    ast.NotIn,
    ast.Subscript,
    ast.Slice,
)

def parse_var_names(expr):
    names = []
    for m in re.finditer('[^\\\](@)', expr):
        i = m.start(1)+1
        prev_s = None
        for j in range(1,len(expr) - i + 1):
            s = expr[i:i+j]
            if not s.isidentifier():
                if prev_s is None:
                    raise ValueError(f"No valid variable name found after @ at index {i}. {expr[i:]=}.")
                names.append(prev_s)
                break
            elif i + j == len(expr):
                names.append(s) # reached the end of string
            else:
                prev_s = s
    return names

def modify_expr(expr, names):
    def backtick_string_replacement(s):
        return f"BACKTICK_QUOTED_STRING_{s.group(0).replace(' ', '_').strip('`')}"

    expr = re.sub('`.*?`', backtick_string_replacement, expr)
    
    for name in names:
        expr = expr.replace(f'@{name}', f'__eval_local_{name}')
    
    parsed_expr = ast.parse(expr)
    if len(parsed_expr.body) != 1:
        raise ValueError(f"expr must consist of a single expression, but {len(parsed_expr.body)=}.")
        
    stmt = parsed_expr.body[0]
    if not isinstance(stmt, ast.Expr):
        raise ValueError(f"expr must be an expression, but {type(stmt)=}.")
    
    class Transformer(ast.NodeTransformer):
        def visit_BoolOp(self, node):
            values = [Transformer().visit(value) for value in node.values] # transform child nodes first
            
            if isinstance(node.op, ast.And):
                op = ast.BitAnd()
            elif isinstance(node.op, ast.Or):
                op = ast.BitOr()
            else:
                raise RuntimeError(f"node.op should be ast.And or ast.Or, but got {node.op=}.")
            
            if len(values) < 2:
                raise RuntimeError(f"Length of node.values should be at least 2, but got {len(values)=}.")
                
            bin_op = ast.BinOp(values.pop(0), op, values.pop(0))
            while len(values) > 0:
                bin_op = ast.BinOp(bin_op, op, values.pop(0))
                
            return bin_op
        
        def visit_UnaryOp(self, node):
            operand = Transformer().visit(node.operand)
            
            op = ast.Invert() if isinstance(node.op, ast.Not) else node.op
            return ast.UnaryOp(op, operand)

    parsed_expr.body[0].value = Transformer().visit(stmt.value)
    
    for node in ast.walk(parsed_expr.body[0].value):
        if not isinstance(node, allowed_nodes):
            raise ValueError(f"{node=} is not allowed in expression.") # Some basic effort to prevent arbitrary code execution.
            
    return ast.unparse(parsed_expr)


import pathlib

import pytest

import hyclib as lib

@pytest.mark.parametrize('filename', [
    ('test_load_config.toml'),
    ('test_load_config.json'),
])
def test_load(filename, pytestconfig):
    data_path = pathlib.Path(pytestconfig.rootdir) / 'test' / 'data'
    config = lib.config.load(data_path / filename)
    assert config == {'a': 1, 'b': 'hi', 'c': True, 'd': [10, 11]}
    
def test_dump(tmp_path):
    config = {'a': 1, 'b': 'hi', 'c': True, 'd': [10, 11]}
    filename = tmp_path / 'config.json'
    lib.config.dump(config, filename)
    loaded_config = lib.config.load(filename)
    assert loaded_config == config
    
def test_expand(pytestconfig):
    filename = 'test_expand_config.toml'
    data_path = pathlib.Path(pytestconfig.rootdir) / 'test' / 'data'
    d = lib.config.load(data_path / filename)
    configs = lib.config.expand(d)
    print(*configs, sep='\n')
    assert configs == [
        {'a': 1, 'b': 'hihi', 'c': [0,1,2,3]},
        {'d': 2, 'e': 'hi'},
        {'d': 3, 'e': 'bye'},
        {'d': 4, 'e': 'beep'},
        {'d': 5, 'e': 'bop'},
        {'a': 1, 'b': 'hi', 'c': 'boop'},
        {'a': 1, 'b': 'bye', 'c': 'boop'},
        {'d': 2, 'e': 'hii'},
        {'d': 5, 'e': 'hii'},
        {'d': 8, 'e': 'hii'},
    ]
    
@pytest.mark.parametrize('filename', [
    ('test_expand_invalid_config_0.toml'),
    ('test_expand_invalid_config_1.toml'),
    ('test_expand_invalid_config_2.toml'),
])
def test_expand_invalid(filename, pytestconfig):
    data_path = pathlib.Path(pytestconfig.rootdir) / 'test' / 'data'
    d = lib.config.load(data_path / filename)
    with pytest.raises(ValueError):
        lib.config.expand(d)
        


from . import exceptions
from . import io
from . import itertools
from . import configurable
from . import pprint
from . import plot
from . import config
from . import argparse
from . import logging
from . import random
from . import warnings
from . import functools
from . import timeit
from . import np
from . import pt
from . import sp
from . import pd
from . import bpd
from . import npf

cfg = config.load_package_config('hyclib')


import torch

import hyclib as lib
import hyclib.configurable as conf

class SubModule(conf.Configurable, torch.nn.Module):
    def __init__(self, a):
        super().__init__()
        self.W = torch.nn.Parameter(torch.ones(2,2))
        self.x = conf.Parameter(dtype=int)
        self.a = a
        
class Module(conf.Configurable, torch.nn.Module):
    def __init__(self, y, z):
        super().__init__()
        self.sub_module = SubModule(conf.Parameter([0,1,2.1]))
        self.y = y
        self.z = z

def test_dict_save_load(tmp_path):
    tmp_path_1 = tmp_path / 'config.json'
    tmp_path_2 = tmp_path / 'state_dict.pth.tar'
    
    m1 = Module(conf.Parameter([1,2]), torch.nn.Parameter(torch.zeros(2,2)))
    lib.io.save_config(tmp_path_1, m1.config_dict())
    torch.save(m1.state_dict(), tmp_path_2)
    
    m2 = Module(conf.Parameter([3,4]), torch.nn.Parameter(10*torch.ones(2,2)))
    config_dict = lib.io.load_config(tmp_path_1)
    state_dict = torch.load(tmp_path_2)
    m2.load_config_dict(config_dict)
    m2.load_state_dict(state_dict)
    
    assert m1.y == m2.y == [1,2]
    assert torch.allclose(m1.z, m2.z)
    assert m1.sub_module.x == m2.sub_module.x == 0
    assert m1.sub_module.a == m2.sub_module.a == [0,1,2.1]
    assert torch.allclose(m1.sub_module.W, m2.sub_module.W)
    
def test_conflicting_attributes():
    m1 = Module(conf.Parameter([1,2]), torch.nn.Parameter(torch.zeros(2,2)))
    m1.y = (10,)
    assert 'y' not in m1.config_dict() and m1.y == (10,)
    m1.y = conf.Parameter(dtype=tuple)
    assert m1.y == tuple()
    m1.y = None
    assert 'y' not in m1.config_dict() and m1.y is None
    m1.sub_module = None
    assert list(m1.config_dict().items()) == []
    m1.y = conf.Parameter(10)
    assert list(m1.config_dict().items()) == [('y',10)]


import functools
import warnings

def parametrized(dec):
    """
    A decorator for decorators.
    Copied from @Dacav's answer here: https://stackoverflow.com/questions/5929107/decorators-with-parameters
    """
    def layer(*args, **kwargs):
        def repl(f):
            print(f, *args, **kwargs)
            return dec(f, *args, **kwargs)
        return repl
    return layer

@parametrized
def deprecated(func, msg=""):
    """This is a decorator which can be used to mark functions
    as deprecated."""
    @functools.wraps(func)
    def new_func(*args, **kwargs):
        warnings.warn(f"Call to deprecated function {func.__name__}. {msg}",
                      category=DeprecationWarning,
                      stacklevel=2)
        return func(*args, **kwargs)
    return new_func



import collections
import re
import logging
import sys

import torch
import numpy as np
import numpy.ma as ma
import pandas as pd

from .. import itertools, timeit
from .. import np as unp
from . import array as ar
from . import parsing

logger = logging.getLogger(__name__)

__all__ = ['NA', 'InternalError', 'DataFrame', 'DataFrameGroupBy', 'concat']

NA = ma.masked

class InternalError(Exception):
    pass
    
def _is_0d_col_idx(obj):
    return isinstance(obj, collections.abc.Hashable) or (isinstance(obj, np.ndarray) and obj.ndim == 0) or (isinstance(obj, torch.Tensor) and obj.ndim == 0)

def _is_1d_col_idx(obj):
    return isinstance(obj, list) or isinstance(obj, slice) or (isinstance(obj, np.ndarray) and obj.ndim == 1) or (isinstance(obj, torch.Tensor) and obj.ndim == 1)

def _is_0d_row_idx(obj):
    return isinstance(obj, int) or (isinstance(obj, np.ndarray) and obj.ndim == 0) or (isinstance(obj, torch.Tensor) and obj.ndim == 0)

def _is_1d_row_idx(obj):
    return isinstance(obj, list) or isinstance(obj, slice) or (isinstance(obj, np.ndarray) and obj.ndim == 1) or (isinstance(obj, torch.Tensor) and obj.ndim == 1)
    
class DataFrame:
    def __init__(self, data, columns=None, copy=True):
        """
        Initialize a DataFrame with the given data and columns.
        Data is stored in the .data attribute and is a dictionary of ar.Array representing the columns.
        If copy is True, constructs a copy of data (arrays and tensor are copied, but for objects, only references are copied).
        Otherwise, only dictionary references (if data is a dictionary) are copied.
        
        When indexing:
            - an element is returned as a 0d torch.Tensor, 0d np.ndarray, or object
            - a column is returned as a 1d torch.Tensor or np.ndarray
            - a row is returned as a dictionary of 0d torch.Tensor, 0d np.ndarray, or object
            - a sub-frame is returned as a DataFrame

        :param data: a 2D np.ndarray/torch.Tensor, a dictionary of (0/1D torch.Tensor/np.ndarray or any other object), or a DataFrame
        :param columns: a list of column names (optional)
        """
        if columns is not None:
            if not (isinstance(columns, collections.abc.Iterable) and isinstance(columns, collections.abc.Sized)) or isinstance(columns, np.ndarray) or isinstance(columns, torch.Tensor):
                raise TypeError(f"columns must be None or be a sized iterable, but {type(columns)=}.")
                
            if len(set(columns)) != len(columns):
                raise ValueError(f"Columns must be unique, but detected duplicates in {columns}.")
        
        if isinstance(data, dict):
            if columns is not None:
                raise ValueError(f"columns must be None when data is a dictionary, but {type(columns)=}.")

            arrs, shapes = [], []
            for v in data.values():
                arr = ar.Array(v)
                if arr.ndim > 1:
                    raise ValueError(f"data must be at most 1D, but {arr.ndim=}.")
                    
                arrs.append(arr)
                shapes.append(arr.shape)
                
            shape = np.broadcast_shapes(*shapes)
            
            if copy:
                data = {k: v.broadcast_to(shape).copy() for k, v in zip(data.keys(), arrs)}
            else:
                data = {k: v.broadcast_to(shape) for k, v in zip(data.keys(), arrs)}
            
        elif isinstance(data, DataFrame):
            if columns is not None:
                raise ValueError(f"column must be None when data is a DataFrame, but {type(columns)=}.")
                
            if copy:
                data = {k: v.copy() for k, v in data._items()}
            else:
                data = data._data.copy() # dictionary shallow copy
            
        elif isinstance(data, np.ndarray) or isinstance(data, torch.Tensor):
            if data.ndim != 2:
                raise ValueError(f"data must be 2D if it is a np.ndarray or torch.Tensor, but {data.ndim=}.")
                
            if columns is None:
                columns = range(data.shape[1])

            if len(columns) != data.shape[1]:
                raise ValueError(f"Number of columns must be the second dim of data, but {len(columns)=}, {data.shape[1]=}.")
            
            data = ar.Array(data)
            if copy:
                data = data.copy()
            data = {column: data[:,i] for i, column in enumerate(columns)}
        
        else:
            raise TypeError(f"data must be a 2D list/np.ndarray/torch.Tensor, a dictionary of (0/1D torch.Tensor/np.ndarray or any other object), or a DataFrame, but got {type(data)}.")
                
        self._data = data
        
    def _validate(self):
        lengths = {}
        for column, data in self._items():
            if not isinstance(ar.Array):
                raise InternalError(f"column data must be ar.Array, but {type(arr)=}.")
            if not data.ndim == 1:
                raise InternalError(f"column data must be 1D, but {data.ndim=}.")
            lengths[column] = len(series)

        if not unp.isconst(lengths.values()):
            raise InternalError(f"each column must have the same length, but {lengths=}.")
        
    @property
    def columns(self):
        return list(self._data.keys())
    
    def __len__(self):
        if len(self._data) == 0:
            return 0
        return len(next(iter(self._data.values())))
    
    @property
    def shape(self):
        return (len(self), len(self._data))
    
    def keys(self):
        yield from self._data.keys()
        
    def _values(self):
        yield from self._data.values()
    
    def values(self):
        for v in self._data.values():
            yield v.data
            
    def _items(self):
        yield from self._data.items()
    
    def items(self):
        for k, v in self._data.items():
            yield k, v.data
        
    def iterrows(self):
        for i in range(len(self)):
            yield i, self[:,i]
     
    @property
    def dtype(self):
        return {k: v.dtype for k, v in self._items()}
    
    @property
    def device(self):
        return {k: v.device if v.is_tensor else None for k, v in self._items()}
    
    @property
    def is_tensor(self):
        return {k: v.is_tensor for k, v in self._items()}
    
    @property
    def nbytes(self):
        return {k: v.nbytes for k, v in self._items()}

    def __getitem__(self, key):
        """
        Index the DataFrame as if indexing a 2D array with arr[:, col_idx][row_idx], where key = (col_idx, row_idx).
        Returns a 0d array if both col_idx and row_idx are 0d, a list of 0d elements if row_idx is 0d,
        a 1d np.ndarray/torch.Tensor if col_idx is 0d, and a dataframe if neither is 0d.
        Note that it is different from indexing a 2D array with arr[col_idx, row_idx] when both col_idx and row_idx are array-like.
        col_idx can only be a slice if it is slice(None) or slice(None, None, step), since it would introduce 
        ambiguity as to whether to interpret slice.start and slice.end as column indices or as column names.
        
        There is no need to worry about making row_idx the same object type and device as column (this would be impossible if
        indexing multiple columns with different object types and devices) - Array.__getitem__ takes care of this issue.
        """
        if isinstance(key, tuple):
            col_idx, row_idx = key
        else:
            col_idx, row_idx = key, slice(None)
            
        if col_idx is Ellipsis:
            col_idx = slice(None)
            
        if row_idx is Ellipsis:
            row_idx = slice(None)
        
        if _is_0d_col_idx(col_idx):
            return self._data[col_idx][row_idx].data
        
        if _is_1d_col_idx(col_idx):
            if isinstance(col_idx, slice):
                if col_idx.start is not None or col_idx.stop is not None:
                    raise ValueError(f"col_idx can only be a slice if both start and end are None, but {col_idx=}")
                    
                col_idx = self.columns[col_idx]

            if _is_0d_row_idx(row_idx):
                return [self._data[k][row_idx].data for k in col_idx]
            
            if _is_1d_row_idx(row_idx):
                return DataFrame({k: self._data[k][row_idx].data for k in col_idx})
            
            raise TypeError(f"Invalid row index {type(row_idx)=}.")

        raise TypeError(f"Invalid column index {type(col_idx)=}.")
        
    def _set_column(self, column, row_idx, arr):
        # NOTE: row_idx = None produces different behavior than row_idx = slice(None)
        # even though both sets all elements of a column
        if row_idx is None:
            self._data[column] = arr.broadcast_to(len(self))
        else:
            if not column in self.columns:
                self._data[column] = ar.Array.empty(
                    (len(self),),
                    is_tensor=arr.is_tensor,
                    dtype=arr.dtype,
                    device=arr.device,
                )
            if not self._data[column].writeable:
                self._data[column] = self._data[column].copy()
            if self._data[column].is_tensor:
                arr = arr.astensor() # allow setting tensor elements with np array
            self._data[column][row_idx] = arr
        
    def __setitem__(self, key, value):
        """
        Set elements of the DataFrame. One can understand it abstractly as doing
        arr.T[col_idx, :][row_idx] = value, where key = (col_idx, row_idx), except
        for two major differences:
            1) In numpy this assignment would not modify the array, but here it would.
            2) In numpy the value is broadcasted to the correct shape by adding leading batch dimensions,
               but here the value is broadcasted to the correct shape by adding trailing batch dimensions.
               
        There is no need to worry about making row_idx the same object type and device as column (this would be impossible if
        indexing multiple columns with different object types and devices) - Array.__setitem__ takes care of this issue.
        """
        if isinstance(key, tuple):
            col_idx, row_idx = key
        else:
            col_idx, row_idx = key, None
            
        if col_idx is Ellipsis:
            col_idx = slice(None)
            
        if row_idx is Ellipsis:
            row_idx = slice(None)
            
        value = ar.Array(value)
        
        if _is_0d_col_idx(col_idx):
            self._set_column(col_idx, row_idx, value)
            
        elif _is_1d_col_idx(col_idx):
            if isinstance(col_idx, slice):
                if col_idx.start is not None or col_idx.stop is not None:
                    raise ValueError(f"col_idx can only be a slice if both start and end are None, but {col_idx=}")
                    
                col_idx = self.columns[col_idx]

            # broadcast value to (len(col_idx), ...) so that we can iterate over columns.
            # this broadcasting different form the usual numpy broadcasting rule in that 
            # batch dimensions here are TRAILING not leading.
            if value.ndim <= 1:
                value = value.broadcast_to(len(col_idx))
            elif value.ndim == 2:
                value = value.broadcast_to((len(col_idx), -1))
            else:
                raise ValueError(f"value must be at most 2D, but {value.ndim=}")
                
            for k, v in zip(col_idx, value):
                self._set_column(k, row_idx, ar.Array(v))
    
    def __repr__(self):
        return repr(self.to_pandas())
        
    def copy(self):
        # Should work similarly to pd.DataFrame.copy(deep=True)
        return DataFrame(self._data, copy=True)
        
    def to(self, device):
        df = self.copy()
        for k, v in df._items():
            df[k] = v.to(device)
        return df
    
    def to_list(self):
        """
        Returns data as a list of columns. Data is copied (but non-array objects are not copied).
        """
        return list(self.copy().values())
    
    def to_dict(self):
        """
        Returns data as a dict of columns. Data is copied (but non-array objects are not copied).
        """
        return self.copy()._data
    
    def to_numpy(self):
        """
        Returns data as a 2D np.ndarray. Data is copied (but non-array objects are not copied).
        """
        return np.stack([v.detach().cpu().numpy() for v in self._values()], axis=1) # np.stack returns copy
    
    def to_torch(self, device=None):
        """
        Returns data as a 2D torch.Tensor, but may raise error if there is a column with a datatype
        that is not supported by pytorch. Data is copied (but non-array objects are not copied).
        """
        return torch.stack([v.astensor().data.to(device) for v in self._values()], dim=1)
    
    def to_pandas(self, copy=True):
        """
        Returns data as a pd.DataFrame.
        """
        return pd.DataFrame({k: v.detach().cpu().numpy() for k, v in self._items()}, copy=copy)

    # def to_html(self, *args, **kwargs):
    #     return self.to_pandas(copy=False).to_html(*args, **kwargs)
    
    def to_html(self, max_rows=None, show_dimensions=False, formatters=None, show_column_info=True):
        """
        chatGPT-assisted code. Slightly slower than using the pd.DataFrame.to_html,
        but allows for more control over display detail, since we don't need to convert
        torch.Tensor to np.ndarray.
        """
        if max_rows is None:
            max_rows = len(self)
        else:
            max_rows = min(max_rows, len(self))
        
        if len(self) <= max_rows:
            row_indices = range(len(self))
        else:
            n_top = max_rows // 2
            n_bottom = max_rows - n_top
            row_indices = list(range(n_top)) + [-1] + list(range(len(self) - n_bottom, len(self)))
        
        html = "<table><thead><tr>"
        html += "<th></th>"
        
        for column in self.columns:
            if show_column_info:
                is_tensor, dtype, device = self._data[column].is_tensor, str(self._data[column].dtype), self._data[column].device
                otype = 'torch' if is_tensor else 'numpy'
                search = re.search('.*\.(.*)', dtype)
                dtype = dtype if search is None else search.group(1)
                device = 'cpu' if device is None else str(device)
                header = [column, otype, dtype, device]
            else:
                header = [column]
            html += f"<th>{'<br/>'.join(header)}</th>"
        html += "</tr></thead><tbody>"
        
        for i in row_indices:
            if i == -1:
                html += "<tr>"
                html += "<td>...</td>" * (len(self.columns) + 1)
                html += "</tr>"
            else:
                html += "<tr>"
                html += f"<td><b>{i}</b></td>"
                for column in self.columns:
                    cell = self[column][i]
                    if formatters is not None and column in formatters:
                        cell = formatters[column](cell)
                    
                    html += f"<td>{cell}</td>"
                html += "</tr>"
        html += "</tbody></table>"
        
        if show_dimensions:
            html += f"<p>{len(self)} rows × {len(self.columns)} columns</p>"
        
        return html
    
    def _to_numpy_numeric(self):
        arrs = []
        for arr in self._values():
            arr = arr.detach().cpu().numpy() # do everything in numpy due to various torch.unique bugs
            if arr.dtype.kind not in 'biufc': # non-numeric
                _, arr = np.unique(arr, return_inverse=True)
            if isinstance(arr, ma.MaskedArray):
                arr = arr.astype(np.result_type(arr.dtype, float)) # promote to at least float so we can turn masked elements to nan
                arr[arr.mask] = np.nan
                arr = arr.data # drop mask, turns ma.MaskedArray into np.ndarray
            arrs.append(arr)
        arr = np.stack(arrs, axis=-1)
        return arr
    
    def info(self):
        return pd.DataFrame([{
            'column': k,
            'n_rows': len(v), # should be same for every column
            'is_tensor': v.is_tensor,
            'dtype': v.dtype,
            'device': v.device,
            'nbytes': v.nbytes,
        } for k, v in self._items()])
    
    def __delitem__(self, columns):
        columns = np.atleast_1d(columns)
        for column in columns:
            del self._data[column]
    
    def drop(self, *columns, copy=False):
        df = DataFrame(self, copy=copy)
        for column in columns:
            del df[column]
        return df
    
    def rename(self, kwargs, copy=False):
        return DataFrame({kwargs.get(k, k): self._data[k] for k in self._data}, copy=copy) # preserves order
    
    def groupby(self, *args, **kwargs):
        # Should works similarly to pd.DataFrame.groupby(by, as_index=False)
        return DataFrameGroupBy(self, *args, **kwargs)
    
    def where(self, condition, level=0):
        var_names = parsing.parse_var_names(condition)
        condition = parsing.modify_expr(condition, var_names)
        frame = sys._getframe(level+1) # see https://github.com/pandas-dev/pandas/blob/main/pandas/core/computation/scope.py
        # add @ variables
        var_dict = {f'__eval_local_{var_name}': frame.f_locals[var_name] for var_name in var_names}
        # add dataframe columns
        var_dict = var_dict | {k: v.data for k, v in self._data.items()}
        # add ability to recognize backtick quoted columns
        var_dict = var_dict | {f"BACKTICK_QUOTED_STRING_{k.replace(' ', '_')}": v.data for k, v in self._data.items()}
        result = eval(condition, var_dict)
        return result
    
    def query(self, condition, level=0):
        return self[:,self.where(condition, level=level+1)]
    
    def merge(self, df, how='inner', on=None, left_on=None, right_on=None, suffixes=('_x', '_y')):
        # if not isinstance(df, DataFrame):
        #     raise TypeError(f"df must be a DataFrame, but {type(df)=}.")
        
        if not how == 'inner':
            raise NotImplementedError()
        
        if on is not None:
            if not (left_on is None and right_on is None):
                raise ValueError(f"left_on and right_on must be None when on is not None, but {left_on=}, {right_on=}.")
            left_on, right_on = on, on
            
        if (left_on is None and right_on is not None) or (left_on is not None and right_on is None):
            raise ValueError(f"left_on and right_on must either be both None or not None, but {left_on=}, {right_on=}.")
        
        if left_on is None: # right_on must also be None
            on = [column for column in self.columns if column in df.columns] # same order as left df
            left_on, right_on = on, on
        else:
            left_on, right_on = np.atleast_1d(left_on), np.atleast_1d(right_on)
            if len(left_on) != len(right_on):
                raise ValueError(f"Length of left_on must be same as length of right_on, but {len(left_on)=}, {len(right_on)=}.")
          
        combined = concat([self[left_on], df[right_on].rename({r: l for l, r in zip(left_on, right_on)})])._to_numpy_numeric()
        larr, rarr = combined[:len(self)], combined[len(self):]
        
        larr, linv, lcount = unp.unique_rows(larr, sorted=False, return_inverse=True, return_counts=True)
        rarr, rinv, rcount = unp.unique_rows(rarr, sorted=False, return_inverse=True, return_counts=True)
        # logger.debug(f'{larr=}')
        # logger.debug(f'{linv=}')
        # logger.debug(f'{lcount=}')
        # logger.debug(f'{rarr=}')
        # logger.debug(f'{rinv=}')
        # logger.debug(f'{rcount=}')
        _, larr_idx, rarr_idx = unp.intersect_rows(larr, rarr, assume_unique=True, return_indices=True)
        # logger.debug(f'{larr_idx=}')
        # logger.debug(f'{rarr_idx=}')
        lmask, rmask = np.isin(linv, larr_idx), np.isin(rinv, rarr_idx)
        # logger.debug(f'{lmask=}')
        # logger.debug(f'{rmask=}')
        lindices, rindices = lmask.nonzero()[0], rmask.nonzero()[0]
        # logger.debug(f'{lindices=}')
        # logger.debug(f'{rindices=}')
        inv_larr_idx, inv_rarr_idx = np.empty(len(larr), dtype=int), np.empty(len(rarr), dtype=int)
        inv_larr_idx[larr_idx] = np.arange(len(larr_idx))
        inv_rarr_idx[rarr_idx] = np.arange(len(rarr_idx))
        # logger.debug(f'{inv_larr_idx=}')
        # logger.debug(f'{inv_rarr_idx=}')
        lindices, rindices = lindices[np.argsort(inv_larr_idx[linv[lmask]])], rindices[np.argsort(inv_rarr_idx[rinv[rmask]])]
        # logger.debug(f'{lindices=}')
        # logger.debug(f'{rindices=}')
        lcount, rcount = lcount[larr_idx], rcount[rarr_idx]
        # logger.debug(f'{lcount=}')
        # logger.debug(f'{rcount=}')
        lindices = np.repeat(lindices, np.repeat(rcount, lcount))
        rindices = unp.repeat(rindices, lcount, chunks=rcount)
        # logger.debug(f'{lindices=}')
        # logger.debug(f'{rindices=}')
        
        left, right = self[:,lindices], df[:,rindices]
        del right[[column for column in left_on if column in right_on]]
        left_rename = {column: f'{column}{suffixes[0]}' for column in left.columns if column in right.columns}
        right_rename = {column: f'{column}{suffixes[1]}' for column in right.columns if column in left.columns}
        
        return concat([left.rename(left_rename), right.rename(right_rename)], axis=1)
    
class DataFrameGroupBy:
    def __init__(self, df, by, sort=True, dropna=True):
        assert isinstance(df, DataFrame)
        self.df = df
        self.by = np.atleast_1d(by)
        self.dropna = dropna
        data = self.df[self.by]
        
        if len(data.columns) == 0:
            raise ValueError("No selected columns.")
            
        arr = self.df[self.by]._to_numpy_numeric() # numeric numpy array representation of selected columns
        _, idx, inv_idx = unp.unique_rows(arr, sorted=sort, return_index=True, return_inverse=True) # faster than np.unique
        # _, idx, inv_idx = np.unique(arr, return_index=True, return_inverse=True, axis=0)
        self._groups = DataFrame({column: self.df[column][idx] for column in self.by})
        self._na = np.stack([arr.isna().detach().cpu().numpy() for arr in self._groups._values()], axis=1).any(axis=1)
        self._row_to_group_idx = ar.Array(inv_idx)
        
    @property
    def groups(self):
        if self.dropna:
            return self._groups[:,~self._na]
        return self._groups
    
    def items(self):
        columns = [column for column in self.df.columns if column not in self.by]
        for idx, group in self._groups.iterrows():
            if self.dropna and self._na[idx]:
                continue
                
            yield tuple(group), self.df[columns, self._row_to_group_idx.data == idx]
        
    def _agg(self, column, func, results):
        if (column, func) in results:
            return results[(column, func)]
        
        _row_to_group_idx = self._row_to_group_idx.like(self.df[column], dtype=False)

        # See scipy.stats.binned_statistic_dd for reference
        if func == 'count':
            result = _row_to_group_idx.bincount()
        
        elif func == 'sum':
            result = _row_to_group_idx.bincount(weights=self.df[column])
        
        elif func == 'mean':
            result = self._agg(column, 'sum', results) / self._agg(column, 'count', results)
            
        elif func == 'var':
            count, mean = self._agg(column, 'count', results), self._agg(column, 'mean', results)
            result = (_row_to_group_idx.bincount(
                weights=(self.df[column] - mean[_row_to_group_idx].data)**2
            ))/(count-1)
            
        elif func == 'std':
            var = self._agg(column, 'var', results)
            result = var**0.5
        
        elif func == 'sem':
            count, std = self._agg(column, 'count', results), self._agg(column, 'std', results)
            result = std / count**0.5
        
        elif func == 'min':
            v = ar.Array(self.df[column])
            i = v.argsort(descending=True)
            result = ar.Array.empty((len(self._groups),), masked=False, is_tensor=v.is_tensor, dtype=v.dtype, device=v.device)
            
            # In pytorch calling __setitem__ with an index tensor that has duplicate elements is undefined and non-deterministic by                 # default. See the documentation on index_put_. In my own experiments, setting use_deterministic_algorithms resolves this               # issue and makes pytorch behave the same way as numpy, but I don't think this behavior is gauranteed.
            if set_enabled := v.is_tensor and not torch.are_deterministic_algorithms_enabled():
                torch.use_deterministic_algorithms(True)
            result[_row_to_group_idx[i]] = v[i] 
            if set_enabled:
                torch.use_deterministic_algorithms(False)
        
        elif func == 'max':
            v = ar.Array(self.df[column])
            i = v.argsort()
            result = ar.Array.empty((len(self._groups),), masked=False, is_tensor=v.is_tensor, dtype=v.dtype, device=v.device)
            
            # In pytorch calling __setitem__ with an index tensor that has duplicate elements is undefined and non-deterministic by                 # default. See the documentation on index_put_. In my own experiments, setting use_deterministic_algorithms resolves this               # issue and makes pytorch behave the same way as numpy, but I don't think this behavior is gauranteed.
            if set_enabled := v.is_tensor and not torch.are_deterministic_algorithms_enabled():
                torch.use_deterministic_algorithms(True)
            result[_row_to_group_idx[i]] = v[i] 
            if set_enabled:
                torch.use_deterministic_algorithms(False)
        
        elif isinstance(func, callable):
            v = ar.Array(self.df[column])
            result = ar.Array.empty((len(self._groups),), masked=False, is_tensor=v.is_tensor, dtype=v.dtype, device=v.device)
            for idx, group in enumerate(self._groups):
                result[idx] = func(v[_row_to_group_idx == idx].data)
        
        else:
            raise ValueError(f"func must be a callable or a string in ['count', 'sum', 'mean', 'var', 'std', 'sem', 'min', 'max'], but got {func}")
        return result
        
    def agg(self, **kwargs):
        data = {k: v for k, v in self.groups.items()}
        results = {}
        for _, v in kwargs.items():
            results[tuple(v)] = self._agg(*v, results)
            
        results = {k: v[~self._na].data if self.dropna else v.data for k, v in results.items()}
        data = data | {k: results[v] for k, v in kwargs.items()}
        
        return DataFrame(data)
        
def concat(dfs, axis=0):
    assert len(dfs) > 0
    all_columns = [df.columns for df in dfs]
    
    if axis == 0:
        if not unp.isconst(all_columns, axis=0).all():
            raise ValueError(f"dfs must have the same columns when axis=0, but {all_columns=}.")
            
        columns = all_columns[0]
        return DataFrame({column: ar.concat([df[column] for df in dfs]) for column in columns})
    
    elif axis == 1:
        columns = itertools.flatten_seq(all_columns)
        if len(set(columns)) != len(columns):
            raise ValueError(f"dfs must have non-overlapping columns when axis=1, but {all_columns=}.")
            
        return DataFrame({k: v for df in dfs for k, v in df.items()})
    
    raise ValueError("axis must be 0 (concatenate along rows) or 1 (concatenate along columns)")


import random
import contextlib

import numpy as np
import torch

@contextlib.contextmanager
def set_seed(seed):
    python_state = random.getstate()
    np_state = np.random.get_state()
    torch_state = torch.random.get_rng_state()
    random.seed(seed)
    np.random.seed(seed)
    torch.random.manual_seed(seed)
    try:
        yield
    finally:
        random.setstate(python_state)
        np.random.set_state(np_state)
        torch.random.set_rng_state(torch_state)


import collections
import functools
import numbers

import torch
import numpy as np
import numpy.ma as ma

from ..np import isconst
from .. import pt

def unpack_args(f):
    @functools.wraps(f)
    def wrapper(self, *args, **kwargs):
        args = [arg.data if isinstance(arg, Array) else arg for arg in args]
        kwargs = {k: v.data if isinstance(v, Array) else v for k, v in kwargs.items()}
        return f(self, *args, **kwargs)
    return wrapper

def concat(arrs, dim=0):
    if len(arrs) == 0:
        raise ValueError(f"arrs must be at least length 1, but {len(arrs)=}.")

    is_tensor = [Array(arr).is_tensor for arr in arrs]
    if not isconst(is_tensor):
        raise TypeError(f"Underlying data of arrs must either all be torch.Tensor or all be np.ndarray, but {is_tensor=}.")

    is_tensor = is_tensor[0]
    arrs = [arr.data for arr in arrs]
    return Array(torch.cat(arrs, dim=dim)) if is_tensor else Array(np.concatenate(arrs, axis=dim))

class Array:
    def __init__(self, data):
        """
        A simple wrapper class that provides some unified operators for np.ndarray and torch.Tensor.
        This class is NOT meant to be used externally.
        Data is NOT copied upon construction (a shallow copy will be constructed if input is a list of objects)
        """
        writeable = True
        if isinstance(data, Array):
            _data, writeable = data.data, data.writeable
        elif isinstance(data, np.ndarray):
            _data, writeable = data, data.flags.writeable
        elif isinstance(data, torch.Tensor):
            _data = data
        elif (isinstance(data, collections.abc.Iterable) and isinstance(data, collections.abc.Sized)) and not isinstance(data, str):
            if all(isinstance(x, numbers.Number) for x in data):
                _data = np.array(data)
            else:
                _data = np.empty((len(data),), dtype=object)
                for i, data_i in enumerate(data):
                    _data[i] = data_i
        else:
            _data = np.array(data)

        self._data = _data
        self._is_tensor = isinstance(_data, torch.Tensor)
        self._device = _data.device if self.is_tensor else torch.device('cpu')
        self._is_masked_array = isinstance(_data, ma.MaskedArray)
        self._writeable = writeable
        
    @property
    def writeable(self):
        return self._writeable
    
    @writeable.setter
    def writeable(self, value):
        # Note: setting writeable to True is probably not a good idea, you probably want to create a copy instead.
        if not self._is_tensor:
            self._data.setflags(write=value)
        self._writeable = value
    
    @classmethod
    def empty(cls, shape, masked=True, is_tensor=False, dtype=None, device=None):
        if not is_tensor:
            if not (device is None or device.type == 'cpu'):
                raise ValueError(f"device must be None or 'cpu' when is_tensor is False, but {device=}")
                
            arr = np.full(shape, np.nan, dtype=dtype)
            
            if masked:
                return cls(ma.array(arr, mask=True, dtype=dtype))
            return cls(arr)
        
        # TODO: return MaskedTensor if masked=True
        return cls(torch.full(shape, torch.nan, device=device))
        
    @property
    def data(self):
        return self._data
    
    @property
    def is_tensor(self):
        return self._is_tensor
    
    @property
    def is_masked_array(self):
        return self._is_masked_array
    
    @property
    def ndim(self):
        return self._data.ndim
    
    @property
    def shape(self):
        return self._data.shape
    
    @property
    def dtype(self):
        return self._data.dtype
    
    @property
    def nbytes(self):
        if self.is_tensor:
            return self.data.storage().nbytes()
        else:
            data = self.data
            while data.base is not None:
                data = data.base
            return data.nbytes
    
    @property
    def device(self):
        return self._device
    
    def __repr__(self):
        return f"Array({repr(self.data)})"
    
    def __str__(self):
        return f"Array({str(self.data)})"
    
    def __len__(self):
        return len(self.data)
    
    @unpack_args
    def __getitem__(self, key):
        if isinstance(key, torch.Tensor) and key.device != self.device:
            key = key.to(self.device)
        # key = Array(key).like(self, dtype=False) # convert to same object type as self.data and move to device if necessary
        arr = Array(self.data[key])
        if self.is_tensor and arr.data.storage().data_ptr() == self.data.storage().data_ptr():
            arr.writeable = self.writeable # same writeable flag since memory is shared
        return arr
    
    @unpack_args
    def __setitem__(self, key, value):
        if isinstance(key, torch.Tensor) and key.device != self.device:
            key = key.to(self.device)
        # key = Array(key).like(self, dtype=False) # convert to same object type as self.data and move to device if necessary
        if not self.writeable:
            raise ValueError("cannot call self.__setitem__ since self.writeable is set to False. Please copy Array first.")
        if self.is_tensor and isinstance(value, torch.Tensor):
            value = value.type(self.dtype) # pytorch bug on mps: tensor assignment is incorrect when value dtype is not the same as data dtype
        self.data[key] = value
        
    def __iter__(self):
        yield from self.data
        
    @unpack_args
    def __add__(self, other):
        return Array(self.data + other)
    
    @unpack_args
    def __sub__(self, other):
        return Array(self.data - other)
    
    @unpack_args
    def __mul__(self, other):
        return Array(self.data * other)
        
    @unpack_args
    def __truediv__(self, other):
        return Array(self.data / other)
    
    @unpack_args
    def __pow__(self, other):
        return Array(self.data ** other)
    
    @unpack_args
    def __radd__(self, other):
        return Array(other + self.data)
    
    @unpack_args
    def __rsub__(self, other):
        return Array(other - self.data)
    
    @unpack_args
    def __rmul__(self, other):
        return Array(other * self.data)
        
    @unpack_args
    def __rtruediv__(self, other):
        return Array(other / self.data)
    
    @unpack_args
    def __rpow__(self, other):
        return Array(other ** self.data)
    
    def astype(self, dtype):
        if self.is_tensor:
            arr = Array(self.data.type(dtype))
            if arr.data.storage().data_ptr() == self.data.storage().data_ptr():
                arr.writeable = self.writeable # same writeable flag since memory is shared
        else:
            arr = Array(data.astype(dtype))
        
        return arr
    
    def asnumpy(self):
        """
        Coerces underlying data to be np.ndarray. Underlying data is not copied.
        """
        if self.is_tensor:
            arr = Array(self.data.numpy())
            arr.writeable = self.writeable # same writeable flag since memory is shared (.numpy() does not copy data)
            return arr
        return self
    
    def astensor(self):
        """
        Coerces underlying data to be torch.Tensor. Underlying data is not copied.
        """
        if self.is_tensor:
            return self
        arr = Array(torch.from_numpy(self.data))
        arr.writeable = self.writeable # same writeable flag since memory is shared (from_numpy() does not copy data)
        return arr

    @unpack_args
    def like(self, other, dtype=True, device=True):
        if self.is_tensor:
            # torch -> torch
            if isinstance(other, torch.Tensor):
                arr = self
                if dtype:
                    arr = arr.astype(other.dtype)
                if device:
                    arr = arr.to(other.device)
                 
            # torch -> numpy
            else:
                if not device and self.device.type != 'cpu':
                    raise ValueError(f"device must True if converting to np.ndarray and self is not on cpu, but {device=}.")
                    
                arr = self.detach().cpu().asnumpy()
                if dtype:
                    arr = arr.astype(other.dtype)
            
        else:
            # numpy -> torch
            if isinstance(other, torch.Tensor):
                arr = self.astensor()
                if dtype:
                    arr = arr.astype(other.dtype)
                if device:
                    arr = arr.to(other.device)
               
            # numpy -> numpy
            else:
                arr = self
                if dtype:
                    arr = arr.astype(other.dtype)
                
        return arr

    def copy(self):
        return Array(self.data.clone()) if self.is_tensor else Array(self.data.copy())
    
    def detach(self):
        if self.is_tensor:
            arr = Array(self.data.detach())
            arr.writeable = self.writeable # same writeable flag since memory is shared
        else:
            arr = self
        return arr
    
    def to(self, device):
        if self.is_tensor:
            arr = Array(self.data.to(device))
            if arr.data.storage().data_ptr() == self.data.storage().data_ptr():
                arr.writeable = self.writeable # same writeable flag since memory is shared
        else:
            arr = self
        return arr
    
    def cpu(self):
        return self.to('cpu')
    
    def numpy(self):
        return self.data.numpy() if self.is_tensor else self.data
    
    def tolist(self):
        return self.data.tolist()
    
    def broadcast_to(self, shape):
        if self.is_tensor:
            arr = Array(self.data.broadcast_to(shape))
            arr.writeable = False # not writeable after broadcasting, just like numpy
        else:
            arr = Array(np.broadcast_to(self.data, shape, subok=True)) # subok=False turns ma.MaskedArray into an np.ndarray
            if arr.is_masked_array:
                mask = np.broadcast_to(self.data.mask, shape) # mask is not automatically broadcasted, so we have to do it ourselves
                arr.data.mask = mask
        return arr
    
    def isna(self):
        if self.is_tensor:
            return Array(self.data.isnan())
        else:
            if isinstance(self.data, np.floating):
                na = np.isnan(self.data)
            else:
                na = np.zeros(self.shape, dtype=bool)
            if self.is_masked_array:
                if isinstance(na, ma.MaskedArray):
                    na = na.data
                na[self.data.mask] = True
            return Array(na)
    
    def unique(self, return_inverse=False, return_counts=False, dim=None):
        if dim is not None:
            raise NotImplementedError(f"unique is not implemented for dim is not None due to ma.MaskedArray bug, but {dim=}.")
        
        if self.is_tensor:
            out = self.data.unique(return_inverse=return_inverse, return_counts=return_counts, dim=dim)
        else:
            out = np.unique(self.data, return_inverse=return_inverse, return_counts=return_counts, axis=dim)
            
        if isinstance(out, tuple):
            return [Array(out_i) for out_i in out]
        return Array(out)
    
    @unpack_args
    def bincount(self, weights=None):
        if self.is_tensor:
            # arr = Array(self.data.bincount(weights=weights))
            ### torch.bincount derivative is not implemented as of torch==1.13.1, so use my custom bincount instead. ###
            arr = Array(pt.bincount(self.data, weights=weights))
        else:
            arr = Array(np.bincount(self.data, weights=weights))
            
        return arr
    
    def argsort(self, descending=False, dim=-1):
        if self.is_tensor:
            return Array(self.data.argsort(descending=descending, dim=dim))
        else:
            arr = Array(np.argsort(self.data, axis=dim))
            if descending:
                arr = arr[::-1]
            return arr